#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#

"""
Test suite for image kernels
"""

from __future__ import division

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-05-28"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""

import time, os, logging
import numpy
import pyopencl, pyopencl.array
import scipy, scipy.misc, scipy.ndimage, pylab
import sys
import unittest
from utilstest import UtilsTest, getLogger, ctx
from test_image_functions import * #for Python implementation of tested functions
from test_image_setup import *
import sift_pyocl as sift
from sift_pyocl.utils import calc_size
logger = getLogger(__file__)
if logger.getEffectiveLevel() <= logging.INFO:
    PROFILE = True
    queue = pyopencl.CommandQueue(ctx, properties=pyopencl.command_queue_properties.PROFILING_ENABLE)
    import pylab
else:
    PROFILE = False
    queue = pyopencl.CommandQueue(ctx)

SHOW_FIGURES = False
PRINT_KEYPOINTS = False
USE_CPU = False
USE_CPP_SIFT = True #use reference cplusplus implementation for descriptors comparison... not valid for (octsize,scale)!=(1,1)



print "working on %s" % ctx.devices[0].name

'''
For Python implementation of tested functions, see "test_image_functions.py"
'''



class test_matching(unittest.TestCase):
    def setUp(self):
        
        kernel_path = os.path.join(os.path.dirname(os.path.abspath(sift.__file__)), ("matching_gpu.cl" if not(USE_CPU) else "matching_cpu.cl"))
        print kernel_path
        kernel_src = open(kernel_path).read()
        self.program = pyopencl.Program(ctx, kernel_src).build() #.build('-D WORKGROUP_SIZE=%s' % wg_size)
        self.wg = (1, 128)



    def tearDown(self):
        self.mat = None
        self.program = None
        
        

    def test_matching(self):
        '''
        tests keypoints matching kernel
        '''    
        image = scipy.misc.lena().astype(numpy.float32)
        try:
            import feature
        except:
            logger.error("WARNING: feature module is not available to compare results with C++ implementation. Matching cannot be tested.")
            feature = None
        
        if (feature != None):
            #get the struct keypoints : (x,y,s,angle,[descriptors])
            sc = feature.SiftAlignment()
            ref_sift = sc.sift(image)
            ref_sift_2 = numpy.recarray((ref_sift.shape),dtype=ref_sift.dtype)
            ref_sift_2[:] = (ref_sift[::-1])
            t0_matching = time.time()
            siftmatch = feature.sift_match(ref_sift, ref_sift_2)
            t1_matching = time.time()
            ref = ref_sift.desc
            
            if (USE_CPU): wg = 1,
            else: wg = 64,
            shape = ref_sift.shape[0]*wg[0],
            
            ratio_th = numpy.float32(0.5329) #sift.cpp : 0.73*0.73
            keypoints_start, keypoints_end = 0, min(ref_sift.shape[0],ref_sift_2.shape[0])
            
            gpu_keypoints1 = pyopencl.array.to_device(queue, ref_sift)
            gpu_keypoints2 = pyopencl.array.to_device(queue, ref_sift_2)
            gpu_matchings = pyopencl.array.zeros(queue, (keypoints_end-keypoints_start,2),dtype=numpy.int32, order="C")
            keypoints_start, keypoints_end = numpy.int32(keypoints_start), numpy.int32(keypoints_end)
            nb_keypoints = numpy.int32(10000)
            counter = pyopencl.array.zeros(queue, (1,1),dtype=numpy.int32, order="C")

            t0 = time.time()
            k1 = self.program.matching(queue, shape, wg,
                    gpu_keypoints1.data, gpu_keypoints2.data, gpu_matchings.data, counter.data,
                    nb_keypoints, ratio_th, keypoints_end, keypoints_end)
            res = gpu_matchings.get()
            cnt = counter.get()
            t1 = time.time()

    #        ref_python, nb_match = my_matching(kp1, kp2, keypoints_start, keypoints_end)
            t2 = time.time()
            
            res_sort = res[numpy.argsort(res[:,1])]
    #        ref_sort = ref[numpy.argsort(ref[:,1])]
            
            print res[0:20]
            print ""
    #        print ref_sort[0:20]
            print("C++ Matching took %.3f ms" %(1000.0*(t1_matching-t0_matching)))
            print("OpenCL: %d match / C++ : %d match" %(cnt,siftmatch.shape[0]))


            #sort to compare added keypoints
            '''
            delta = abs(res_sort-ref_sort).max()
            self.assert_(delta == 0, "delta=%s" % (delta)) #integers
            logger.info("delta=%s" % delta)
            '''

            if PROFILE:
                logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
                logger.info("Matching took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))
                
                
            
            

def test_suite_matching():
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_matching("test_matching"))
    return testSuite

if __name__ == '__main__':
    mysuite = test_suite_matching()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        sys.exit(1)




#!/usr/bin/env python
import numpy
def normalize_image(img):
    maxi = numpy.float32(img.max())
    mini = numpy.float32(img.min())
    return numpy.ascontiguousarray(numpy.float32(255) * (img - mini) / (maxi - mini), dtype=numpy.float32)

def shrink(img, xs, ys):
    return img[0::ys, 0::xs]



def my_gradient(mat):
    """
    numpy implementation of gradient :
    "The gradient is computed using central differences in the interior and first differences at the boundaries. The returned gradient hence has the same shape as the input array."
    NOTE:
        -with numpy.gradient, the amplitude is twice smaller than in SIFT.cpp, therefore we multiply the amplitude by two
    """
    g = numpy.gradient(mat)
    return 2.0*numpy.sqrt(g[0]**2+g[1]**2), numpy.arctan2(g[0],g[1]) #sift.cpp puts a "-" here
    
    
    
    
    
    
def my_local_maxmin(DOGS,thresh,border_dist,octsize,EdgeThresh0,EdgeThresh,nb_keypoints,s,dog_width,dog_height):
    """
    a python implementation of 3x3 maximum (positive values) or minimum (negative or null values) detection
    an extremum candidate "val" has to be greater than 0.8*thresh
    The three DoG have the same size.
    """
    output = -numpy.ones((nb_keypoints,4),dtype=numpy.float32) #for invalid keypoints
    
    dog_prev = DOGS[s-1]
    dog = DOGS[s]
    dog_next = DOGS[s+1]
    counter = 0
    
    for j in range(border_dist,dog_width - border_dist):
        for i in range(border_dist,dog_height - border_dist):
            val = dog[i,j]
            if (numpy.abs(val) > 0.8*thresh): #keypoints refinement: eliminating low-contrast points
                if (is_maxmin(dog_prev,dog,dog_next,val,i,j,octsize,EdgeThresh0,EdgeThresh) != 0):
                    output[counter,0]=val
                    output[counter,1]=i
                    output[counter,2]=j
                    output[counter,3]=numpy.float32(s)
                    counter+=1              
    return output, counter
    
    
def is_maxmin(dog_prev,dog,dog_next,val,i0,j0,octsize,EdgeThresh0,EdgeThresh):
    """
    return 1 iff mat[i0,j0] is a local (3x3) maximum
    return -1 iff mat[i0,j0] is a local (3x3) minimum
    return 0 by default (neither maximum nor minimum, or value on an edge)
     * Assumes that we are not on the edges, i.e border_dist >= 2 above
    """
    ismax = 0
    ismin = 0
    res = 0
    if (val > 0.0): ismax = 1
    else: ismin = 1
    for j in range(j0-1,j0+1+1):
        for i in range(i0-1,i0+1+1):
            if (ismax == 1):
                if (dog_prev[i,j] > val or dog[i,j] > val or dog_next[i,j] > val): ismax = 0
            if (ismin == 1):
                if (dog_prev[i,j] < val or dog[i,j] < val or dog_next[i,j] < val): ismin = 0
    
    if (ismax == 1): res =  1 
    if (ismin == 1): res = -1
    
    #keypoint refinement: eliminating points at edges
    H00 = dog[i0-1,j0] - 2.0 * dog[i0,j0] + dog[i0+1,j0]
    H11 = dog[i0,j0-1]- 2.0 * dog[i0,j0] + dog[i0,j0+1]
    H01 = ( (dog[i0+1,j0+1] - dog[i0+1,j0-1])
        - (dog[i0-1,j0+1] - dog[i0-1,j0-1]) ) / 4.0

    det = H00 * H11 - H01 * H01
    trace = H00 + H11

    if (octsize <= 1):
        thr = EdgeThresh0
    else:
        thr = EdgeThresh
    if (det < thr * trace * trace):
        res = 0
        
    return res
    









def my_interp_keypoint(DOGS, s, r, c,movesRemain,peakthresh,width,height):
    ''''
     A Python implementation of SIFT "InterpKeyPoints"
     (s,r,c) : coords of the processed keypoint in the scale space
     WARNING: replace "1.6" by "InitSigma" if InitSigma has not its default value 
     The recursive calls were replaced by a loop.
    '''
    if (r == -1): return (-1,-1,-1,-1)
    dog_prev = DOGS[s-1]
    dog = DOGS[s]
    dog_next = DOGS[s+1]
    newr = r
    newc = c
    loop = 1
    movesRemain = 5
    while (loop == 1):
        
        r0, c0 = newr, newc
        x,peakval = fit_quadratic(dog_prev,dog,dog_next, newr, newc)
        
        if (x[1] > 0.6 and newr < height - 3):
            newr+=1
        elif (x[1] < -0.6 and newr > 3):
            newr-=1
        if (x[2] > 0.6 and newc < width - 3):
            newc+=1
        elif (x[2] < -0.6 and newc > 3):
            newc-=1
 
        #loop test
        if (movesRemain > 0  and  (newr != r or newc != c)):
            movesRemain-=1
        else:
            loop = 0
            

    if (abs(x[0]) <  1.5 and abs(x[1]) <  1.5 and abs(x[2]) <  1.5 and abs(peakval) > peakthresh):
        ki = numpy.zeros(4,dtype=numpy.float32)
        ki[0] = peakval
        ki[1] = r0 + x[1]
        ki[2] = c0 + x[2]
        ki[3] = 1.6 * 2.0**((float(s) + x[0]) / 3.0) #3.0 is "par.Scales" 
    else:
        ki = (-1,-1,-1,-1)
    
    return ki #our interpolated keypoint


def fit_quadratic(dog_prev,dog,dog_next, r, c):
    '''
    quadratic interpolation arround the keypoint (s,r,c)
    '''

    #gradient
    g = numpy.zeros(3,dtype=numpy.float32)
    g[0] = (dog_next[r,c] - dog_prev[r,c]) / 2.0
    g[1] = (dog[r+1,c] - dog[r-1,c]) / 2.0
    g[2] = (dog[r,c+1] - dog[r,c-1]) / 2.0
    #hessian
    H = numpy.zeros((3,3)).astype(numpy.float32)
    H[0][0] = dog_prev[r,c]   - 2.0 * dog[r,c] + dog_next[r,c]
    H[1][1] = dog[r-1,c] - 2.0 * dog[r,c] + dog[r+1,c]
    H[2][2] = dog[r,c-1] - 2.0 * dog[r,c] + dog[r,c+1]
    H[0][1] = H[1][0] = ( (dog_next[r+1,c] - dog_next[r-1,c])
                         - (dog_prev[r+1,c] - dog_prev[r-1,c]) ) / 4.0


    H[0][2] = H[2][0] = ( (dog_next[r,c+1] - dog_next[r,c-1])
                     - (dog_prev[r,c+1] - dog_prev[r,c-1]) ) / 4.0

    H[1][2] = H[2][1]= ( (dog[r+1,c+1] - dog[r+1,c-1])
                     - (dog[r-1,c+1] - dog[r-1,c-1]) ) / 4.0
             
    x = -numpy.dot(numpy.linalg.inv(H),g) #extremum position
    peakval = dog[r,c] + 0.5 * (x[0]*g[0]+x[1]*g[1]+x[2]*g[2])
    
    return x, peakval
    
    
    
    

    
    
def my_orientation(keypoints, nb_keypoints, keypoints_start, keypoints_end, grad, ori, octsize, orisigma):
    '''
    Python implementation of orientation assignment
    '''
    
    counter = keypoints_end #actual_nb_keypoints
    hist = numpy.zeros(36,dtype=numpy.float32)
    rows,cols = grad.shape
    for index,k in enumerate(keypoints[keypoints_start:keypoints_end]):
        if (k[1] != -1.0):
            hist = hist*0 #do not forget this memset at each loop...
            row = numpy.int32(k[1]+0.5)
            col = numpy.int32(k[2]+0.5)
            sigma = orisigma * k[3]
            radius = numpy.int32(sigma * 3.0)
            rmin = max(0,row-radius)
            cmin = max(0,col-radius)
            rmax = min(row+radius,rows-2)
            cmax = min(col+radius,cols-2)
            radius2 = numpy.float32(radius * radius)
            sigma2 = 2.0*sigma*sigma
            #print rmin, rmax, cmin, cmax
            
            for r in range(rmin,rmax+1):
                for c in range(cmin,cmax+1):
                    gval = grad[r,c]
                    distsq = (r-k[1])*(r-k[1]) + (c-k[2])*(c-k[2])
                    
                    if (gval > 0.0  and  distsq < radius2 + 0.5):
                        weight = numpy.exp(- distsq / sigma2)
                        angle = ori[r,c]
                        mybin = numpy.int32((36 * (angle + numpy.pi + 0.001) / (2.0 * numpy.pi)))
                        if (mybin >= 0 and mybin <= 36):
                            mybin = min(mybin, 36 - 1)
                            hist[mybin] += weight * gval

            for i in range(0,6):
                hist = smooth_histogram(hist)
                
            maxval = hist.max()
            argmax = hist.argmax()

            if argmax == 0: prev = 35
            else: prev = argmax -1
            if argmax == 35: next = 0
            else: next = argmax +1
            if (maxval < 0.0):
                hist[prev] = -hist[prev] 
                maxval = -maxval
                hist[next] = -hist[next]

            interp = 0.5 * (hist[prev] - hist[next]) / (hist[prev] - 2.0 * maxval + hist[next])
            angle = 2.0 * numpy.pi * (argmax + 0.5 + interp) / 36 - numpy.pi        
            k[0] = k[2] *octsize
            k[1] = k[1] *octsize
            k[2] = k[3] *octsize
            k[3] = angle
            keypoints[index] = k 
            
            k2 = numpy.zeros(4,dtype=numpy.float32)
            k2[0] = k[0]; k2[1] = k[1]; k2[2] = k[2]; k2[3] = 0.0
            for i in range(0,36):
                if i == 0: prev = 35
                else: prev = i-1
                if i == 35: next = 0
                else: next = i+1
                if (hist[i] > hist[prev]  and  hist[i] > hist[next]  and hist[i] >= 0.8 * maxval and i != argmax):
                    if (hist[i] < 0.0):
                        hist[prev] = -hist[prev]
                        hist[i] = -hist[i]
                        hist[next] = -hist[next]
                    if (hist[i] >= hist[prev]  and  hist[i] >= hist[next]): 
                         interp = 0.5 * (hist[prev] - hist[next]) / (hist[prev] - 2.0 * hist[i] + hist[next])
                
                    angle = 2.0 * numpy.pi * (i + 0.5 + interp) / 36 - numpy.pi
                    if (angle >= -numpy.pi  and  angle <= numpy.pi):
                        k2[3] = angle
                        if (counter < nb_keypoints):
                            keypoints[counter] = k2
                            counter+=1
                
            #end of additional keypoints creation
        #end of "if valid keypoint"
    #end of loop
    return keypoints, counter




def smooth_histogram(hist):
    prev = hist[35]
    for i in range(0,36):
        temp = hist[i]
        if (i+1 == 36): idx = 0
        else: idx = i+1
        hist[i] = ( prev + hist[i] + hist[idx] ) / 3.0
        prev = temp
    return hist









def my_descriptor(keypoints, grad, orim, octsize, keypoints_start, keypoints_end):
    '''
    Python implementation of keypoint descriptor computation
    '''
    #a descriptor is a 128-vector (4,4,8) ; we need keypoints_end-keypoints_start+1  descriptors
    descriptors = numpy.zeros((keypoints_end-keypoints_start,4,4,8),dtype=numpy.float32)
    for index,k in enumerate(keypoints[keypoints_start:keypoints_end]):
        if (k[1] != -1.0):
            irow, icol = int(k[1]/octsize + 0.5), int(k[0]/octsize + 0.5)
            sine, cosine = numpy.sin(k[3]), numpy.cos(k[3])
            spacing = k[2]/octsize * 3
            iradius = int((1.414 * spacing * (5) / 2.0) + 0.5)
            for i in range(-iradius,iradius+1):
                for j in range(-iradius,iradius+1):
                    (rx, cx) = (numpy.dot(numpy.array([[cosine,-sine],[sine,cosine]]),numpy.array([i,j]))
                                -numpy.array([k[1]/octsize-irow,k[0]/octsize-icol]))/spacing + 1.5
                        
                    if (rx > -1.0 and rx < 4.0 and cx > -1.0 and cx < 4.0
                         and (irow +i) >= 0  and (irow +i) < grad.shape[0] and (icol+j) >= 0 and (icol+j) < grad.shape[1]):

                        mag = grad[int(irow+i),int(icol+j)] * numpy.exp(- ((rx - 1.5)**2 + (cx - 1.5)**2) / 8.0)
                        ori = orim[int(irow+i),int(icol+j)] -  k[3]

                        while (ori > 2.0*numpy.pi): ori -= 2.0*numpy.pi
                        while (ori < 0.0): ori += 2.0*numpy.pi
                    
                        oval = 8 * ori / (2.0*numpy.pi)
                        
                        ri = int(rx if (rx >= 0.0) else rx-1.0)
                        ci = int(cx if (cx >= 0.0) else cx -1.0)
                        oi = int(oval if (oval >= 0.0) else oval - 1.0)
                        rfrac, cfrac, ofrac = rx - ri,  cx - ci, oval - oi            
                        
                        if (ri >= -1  and  ri < 4  and oi >=  0  and  oi <= 8  and rfrac >= 0.0 and rfrac <= 1.0):
                            for r in range(0,2):
                                rindex = ri + r
                                if (rindex >=0 and rindex < 4):
                                    rweight = mag * (1.0 -rfrac if (r == 0) else rfrac)
                                    for c in range(0,2):
                                        cindex = ci + c
                                        if (cindex >=0 and cindex < 4):
                                            cweight = rweight * (1.0 - cfrac if (c == 0) else cfrac)
                                            for orr in range(0,2):
                                                oindex = oi + orr
                                                if (oindex >= 8): oindex = 0
                                                descriptors[index][rindex][cindex][oindex] += cweight * (1.0 - ofrac if (orr == 0) else ofrac)
                                        # end "valid cindex"
                                # end "valid rindex"
                        # end "sample in boundaries"
                # end "j loop"
            # end "i loop"
        # end "valid keypoint"
    #end loop in keypoints
   
    #unwrap and normalize the 128-vector
    descriptors = descriptors.reshape(keypoints_end-keypoints_start,128)
    
    
    for i in range(0,keypoints_end-keypoints_start): descriptors[i] = normalize(descriptors[i])
     
    #threshold to 0.2 like in sift.cpp
    changed = 0
    for i in range(0,keypoints_end-keypoints_start):
        idx = descriptors[i] > 0.2
        if (idx.shape[0] != 0):
            (descriptors[i])[idx] = 0.2
            changed = 1
   
    #if values were actually threshold, we have to normalize again
    if (changed == 1):
       for i in range(0,keypoints_end-keypoints_start):
           descriptors[i] = normalize(descriptors[i])
    
    #cast to "unsigned char"
    descriptors = 512*descriptors
    for i in range(0,keypoints_end-keypoints_start): (descriptors[i])[255<=descriptors[i]] = 255
    
    descriptors = descriptors.astype(numpy.uint8)
    
    return descriptors
    
        


def normalize(vec):
    return (vec/numpy.linalg.norm(vec) if numpy.linalg.norm(vec) != 0 else 0)




def my_matching(keypoints1, keypoints2, start, end, ratio_th=0.5329):
    '''
    Python implementation of SIFT keypoints matching
    '''
    counter = 0
    matchings = numpy.zeros((end-start,2),dtype=numpy.uint32)
    for i, desc1 in enumerate(keypoints1):#FIXME: keypoints1.desc.... idem below
        ratio, match = check_for_match(desc1, keypoints2)
        if (ratio < ratio_th and i <= match):
            matchings[counter] = i, match
            counter += 1
    return matchings, counter  


def check_for_match(desc1, keypoints2):
    '''
    check if the descriptor "desc1" has matches in the list "keypoints2"
    '''
    current_min = 0
    dist1 = dist2 = 1000000000000.0
    for j, desc2 in enumerate(keypoints2):
        dst = l1_distance(desc1,desc2)
        if (dst < dist1):
		    dist2 = dist1
		    dist1 = dst
		    current_min = j
        elif (dst < dist2):
		    dist2 = dst;

    return dist1/dist2, current_min



def l1_distance(desc1, desc2):
    '''
    L1 distance between two vectors
    '''
    return abs(desc1-desc2).sum()













def keypoints_compare(ref,res):
    '''
    When using atomic instructions in kernels, the resulting keypoints are not in the same order as Python implementation
    '''
    res_c = res[(res[:,0].argsort(axis=0)),0]
    ref_c = ref[(ref[:,0].argsort(axis=0)),0]
    res_r = res[(res[:,1].argsort(axis=0)),1]
    ref_r = ref[(ref[:,1].argsort(axis=0)),1]
    res_s = res[(res[:,2].argsort(axis=0)),2]
    ref_s = ref[(ref[:,2].argsort(axis=0)),2]
    res_angle = res[(res[:,3].argsort(axis=0)),3]
    ref_angle = ref[(ref[:,3].argsort(axis=0)),3]
    
    return abs(res_c - ref_c).max(), abs(res_r - ref_r).max(), abs(res_s - ref_s).max(), abs(res_angle - ref_angle).max()


def descriptors_compare(ref,res):
    #count null descriptors in "ref" (take care of the order in the arguments : (ref, res) and not (res, ref))
    nulldesc = 0
    for descriptor2 in res:
        if abs(descriptor2).sum() == 0: nulldesc+=1
    #count descriptors that are (exactly) equal
    match = 0
    delta = 0
    for descriptor in ref:
        for descriptor2 in res:
            delta = abs(descriptor-descriptor2).sum()
            if delta == 0: match+=1
            
    return match,nulldesc
        
def check_for_doubles(res):
    #check for descriptors that appear more than one time in the list
    doubles = 0
    cnt = numpy.zeros(res.shape[0])
    for idx,desc in enumerate(res):
        for idx2,desc2 in enumerate(res):
            if abs(desc-desc2).sum() == 0 and idx != idx2 and (cnt[idx] == 0 or cnt[idx2] == 0):
                cnt[idx] = 1
                cnt[idx2] = 1
                doubles+=1
    return doubles



def my_compact(keypoints,nbkeypoints):
    '''
    Reference compacting
    '''
    output = -numpy.ones_like(keypoints)
    idx = numpy.where(keypoints[:,1]!=-1)[0]
    length = idx.size
    output[:length,0] = keypoints[idx,0]
    output[:length,1] = keypoints[idx,1]
    output[:length,2] = keypoints[idx,2]
    output[:length,3] = keypoints[idx,3]
    return output, length
 
'''    
    
    
  function KahanSum(input)
    var sum = 0.0
    var c = 0.0
    for i = 1 to input.length do
        y = input[i] - c  
        t = sum + y
        c = (t - sum) - y
        sum = t
    return sum  
    

    
'''   
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    




#!/usr/bin/env python
import numpy, scipy.ndimage, scipy.misc
from test_image_functions import *
from test_algebra import *
from math import ceil
'''
Unit tests become more and more difficult as we progress in the global SIFT algorithm
For a better code visibility, the setups required by kernels will be put here
'''


def my_blur(img, sigma):
    ksize = int(ceil(8 * sigma + 1))
    if (ksize % 2 == 0):
        ksize += 1
    x = numpy.arange(ksize) - (ksize - 1.0) / 2.0
    gaussian = numpy.exp(-(x / sigma) ** 2 / 2.0).astype(numpy.float32)
    gaussian /= gaussian.sum(dtype=numpy.float32)
    tmp1 = scipy.ndimage.filters.convolve1d(img, gaussian, axis= -1, mode="reflect")
    return scipy.ndimage.filters.convolve1d(tmp1, gaussian, axis=0, mode="reflect")



def local_maxmin_setup():
    border_dist = numpy.int32(5) #SIFT
    peakthresh = numpy.float32(255.0 * 0.04 / 3.0) #SIFT uses 255.0 * 0.04 / 3.0
    EdgeThresh = numpy.float32(0.06) #SIFT
    EdgeThresh0 = numpy.float32(0.08) #SIFT
    octsize = numpy.int32(2) #initially 1, then twiced at each new octave
    scale = numpy.int32(1) #1,2 or 3
    nb_keypoints = 1000 #constant size !
    doubleimsize = 0 #par.DoubleImSize = 0 by default

    l2 = scipy.misc.lena().astype(numpy.float32)#[100:250,100:250] #use a part of the image to fasten tests
    l2 = numpy.ascontiguousarray(l2[0:507, 0:209]);
    #l2 = scipy.misc.imread("../aerial.tiff").astype(numpy.float32)
    l = normalize_image(l2) #do not forget to normalize the image if you want to compare with sift.cpp
    for octave_cnt in range(1, int(numpy.log2(octsize)) + 1 + 1):


        width = numpy.int32(l.shape[1])
        height = numpy.int32(l.shape[0])

        #Blurs and DoGs pre-allocating
        g = (numpy.zeros(6 * height * width).astype(numpy.float32)).reshape(6, height, width) #vector of 6 blurs
        DOGS = numpy.zeros((5, height, width), dtype=numpy.float32) #vector of 5 DoGs
        g[0, :, :] = numpy.copy(l)
        '''
        sift.cpp pre-process
        '''
        if (octave_cnt == 1):
            initsigma = 1.6
            if (doubleimsize): cursigma = 1.0
            else: cursigma = 0.5
            #Convolving initial image to achieve std = initsigma = 1.6
            if (initsigma > cursigma):
                sigma = numpy.sqrt(initsigma ** 2 - cursigma ** 2)
                g[0, :, :] = my_blur(l, sigma)
        else:
            g[0, :, :] = numpy.copy(l)
        '''
        Blurs and DoGs
        '''
        sigmaratio = 2 ** (1 / 3.0) #sift.cpp
        #sift.cpp : for a given "i", we have : increase = initsigma*(sigmaratio)^(i-1)*sqrt(sigmaratio**2 -1)
        for i in range(1, 6):
            sigma = initsigma * (sigmaratio) ** (i - 1.0) * numpy.sqrt(sigmaratio ** 2 - 1.0) #sift.cpp "increase"
            g[i] = my_blur(g[i - 1], sigma) #blur[i]

        for s in range(1, 6): DOGS[s - 1] = -(g[s] - g[s - 1]) #DoG[s-1]

        if (octsize > 1): #if a higher octave is required, we have to sample Blur[3]
            l = shrink(g[3], 2, 2)
    #end for

    #print("[Octave %s] printing blur 2" %(int(numpy.log2(octsize))+1))
    #print g[2,0:10,0:10]
    #print("[Octave %s] printing dog 2" %(int(numpy.log2(octsize))+1))
    #print DOGS[2,0:10,0:10]

    return border_dist, peakthresh, EdgeThresh, EdgeThresh0, octsize, scale, nb_keypoints, width, height, DOGS, g




def interpolation_setup():
    '''
    Provides the values required by "test_interpolation"
    Previous step: local extrema detection - we got a vector of keypoints to be interpolated
    '''

    border_dist, peakthresh, EdgeThresh, EdgeThresh0, octsize, s, nb_keypoints, width, height, DOGS, g = local_maxmin_setup()

    nb_keypoints = numpy.int32(nb_keypoints)

    #Assumes that local_maxmin is working so that we can use Python's "my_local_maxmin" instead of the kernel
    keypoints_prev, actual_nb_keypoints = my_local_maxmin(DOGS, peakthresh, border_dist, octsize,
        EdgeThresh0, EdgeThresh, nb_keypoints, s, width, height)

    return border_dist, peakthresh, EdgeThresh, EdgeThresh0, octsize, nb_keypoints, actual_nb_keypoints, width, height, DOGS, s, keypoints_prev, g[s]




def orientation_setup():
    '''
    Provides the values required by "test_orientation"
    Previous step: interpolation - we got a vector of valid keypoints
    '''
    border_dist, peakthresh, EdgeThresh, EdgeThresh0, octsize, nb_keypoints, actual_nb_keypoints, width, height, DOGS, s, keypoints_prev, blur = interpolation_setup()

    #actual_nb_keypoints = numpy.int32(len((keypoints_prev[:,0])[keypoints_prev[:,1] != -1]))
    ref = numpy.copy(keypoints_prev)
    #There are actually less than "actual_nb_keypoints" keypoints ("holes" in the vector), but we can use it as a boundary
    for i, k in enumerate(ref[:actual_nb_keypoints, :]):
        ref[i] = my_interp_keypoint(DOGS, s, k[1], k[2], 5, peakthresh, width, height)

    grad, ori = my_gradient(blur) #gradient is applied on blur[s]
   # ref, actual_nb_keypoints = my_compact(ref,nb_keypoints)

    return ref, nb_keypoints, actual_nb_keypoints, grad, ori, octsize



def descriptor_setup():
    '''
    Provides the values required by "test_descriptor"
    Previous step: orientation - we got a vector of keypoints with an orientation, and several additional keypoints
    '''
    keypoints, nb_keypoints, actual_nb_keypoints, grad, ori, octsize = orientation_setup()
    orisigma = numpy.float32(1.5) #SIFT
    keypoints_start = numpy.int32(0)
    keypoints_end = actual_nb_keypoints #numpy.int32(actual_nb_keypoints)
    ref, updated_nb_keypoints = my_orientation(keypoints, nb_keypoints, keypoints_start, keypoints_end, grad, ori, octsize, orisigma)

    return ref, nb_keypoints, updated_nb_keypoints, grad, ori, octsize





def matching_setup():
    '''
    Provides the values required by "test_matching"
    Previous step: descriptors - we got a vector of 128-values descriptors
    '''
    keypoints, nb_keypoints, actual_nb_keypoints, grad, ori, octsize = descriptor_setup()
    keypoints, actual_nb_keypoints = my_compact(numpy.copy(keypoints), nb_keypoints)
    keypoints_start, keypoints_end = 0, actual_nb_keypoints
    desc = my_descriptor(keypoints, grad, ori, octsize, keypoints_start, keypoints_end)
    #keypoints with their descriptors
    #FIXME: structure including keypoint (float32) and descriptors (uint8)
    kp1 = desc
    kp2 = numpy.ascontiguousarray(desc[::-1])
    return kp1, kp2, nb_keypoints, actual_nb_keypoints









class Enum(dict):
    """
    Simple class half way between a dict and a class, behaving as an enum
    """
    def __getattr__(self, name):
        if name in self:
            return self[name]
        raise AttributeError

par = Enum(OctaveMax=100000,
            DoubleImSize=0,
            order=3,
            InitSigma=1.6,
            BorderDist=5,
            Scales=3,
            PeakThresh=255.0 * 0.04 / 3.0,
            EdgeThresh=0.06,
            EdgeThresh1=0.08,
#To detect an edge response, we require the ratio of smallest
#to largest principle curvatures of the DOG function
#(eigenvalues of the Hessian) to be below a threshold.  For
#efficiency, we use Harris' idea of requiring the determinant to
#be above par.EdgeThresh times the squared trace, as for eigenvalues
#A and B, det = AB, trace = A+B.  So if A = 10B, then det = 10B**2,
#and trace**2 = (11B)**2 = 121B**2, so par.EdgeThresh = 10/121 =
#0.08 to require ratio of eigenvalues less than 10.
            OriBins=36,
            OriSigma=1.5,
            OriHistThresh=0.8,
            MaxIndexVal=0.2,
            MagFactor=3,
            IndexSigma=1.0,
            IgnoreGradSign=0,
            MatchRatio=0.73,
            MatchXradius=1000000.0,
            MatchYradius=1000000.0,
            noncorrectlylocalized=0)





#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#

"""
Automatic selection of OpenCL devices 
"""

from __future__ import division, print_function, with_statement

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-11-21"
__status__ = "beta"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""

import os, logging
logger = logging.getLogger("sift.opencl")

try:
    import pyopencl, pyopencl.array
except ImportError:
    logger.warning("Unable to import pyOpenCl. Please install it from: http://pypi.python.org/pypi/pyopencl")
    pyopencl = None

FLOP_PER_CORE = { "GPU": 64, # GPU, Fermi at least perform 64 flops per cycle/multicore, G80 were at 24 or 48 ...
                  "CPU": 4,  # CPU, at least intel's have 4 operation per cycle
                  "ACC": 8}  # ACC: the Xeon-phi (MIC) appears to be able to process 8 Flops per hyperthreaded-core
NVIDIA_FLOP_PER_CORE = {(1, 0): 24,  # Guessed !
                         (1, 1): 24,  # Measured on G98 [Quadro NVS 295]
                         (1, 2): 24,  # Guessed !
                         (1, 3): 24,  # measured on a GT285 (GT200)
                         (2, 0): 64,  # Measured on a 580 (GF110)
                         (2, 1): 96,  # Measured on Quadro2000 GF106GL
                         (3, 0): 384, # Guessed!
                         (3, 5): 384, # Measured on K20
                         (5, 0): 256} # Maxwell 4 warps/SM 2 flops/ CU
AMD_FLOP_PER_CORE = 160  # Measured on a M7820 10 core, 700MHz 1120GFlops

class Device(object):
    """
    Simple class that contains the structure of an OpenCL device
    """
    def __init__(self, name="None", dtype=None, version=None, driver_version=None,
                 extensions="", memory=None, available=None,
                 cores=None, frequency=None, flop_core=None, idx=0, workgroup=1):
        """
        Simple container with some important data for the OpenCL device description:
        
        @param name: name of the device
        @param dtype: device type: CPU/GPU/ACC...
        @param version: driver version
        @param driver_version: 
        @param extensions: List of opencl extensions
        @param memory: maximum memory available on the device
        @param available: is the device desactivated or not 
        @param cores: number of SM/cores
        @param frequency: frequency of the device
        @param flop_cores: Flopating Point operation per core per cycle
        @param idx: index of the device within the platform
        @param workgroup: max workgroup size
        """
        self.name = name.strip()
        self.type = dtype
        self.version = version
        self.driver_version = driver_version
        self.extensions = extensions.split()
        self.memory = memory
        self.available = available
        self.cores = cores
        self.frequency = frequency
        self.id = idx
        self.max_work_group_size = workgroup
        if not flop_core:
            flop_core = FLOP_PER_CORE.get(dtype, 1)
        if cores and frequency:
            self.flops = cores * frequency * flop_core
        else:
            self.flops = flop_core


    def __repr__(self):
        return "%s" % self.name

    def pretty_print(self):
        """
        Complete device description
        
        @return: string
        """
        lst = ["Name\t\t:\t%s" % self.name,
               "Type\t\t:\t%s" % self.type,
               "Memory\t\t:\t%.3f MB" % (self.memory / 2.0 ** 20),
               "Cores\t\t:\t%s CU" % self.cores,
               "Frequency\t:\t%s MHz"%self.frequency,
               "Speed\t\t:\t%.3f GFLOPS" % (self.flops / 1000.),
               "Version\t\t:\t%s" % self.version,
               "Available\t:\t%s" % self.available]
        return os.linesep.join(lst)


class Platform(object):
    """
    Simple class that contains the structure of an OpenCL platform
    """
    def __init__(self, name="None", vendor="None", version=None, extensions=None, idx=0):
        self.name = name.strip()
        self.vendor = vendor.strip()
        self.version = version
        self.extensions = extensions.split()
        self.devices = []
        self.id = idx

    def __repr__(self):
        return "%s" % self.name

    def add_device(self, device):
        self.devices.append(device)

    def get_device(self, key):
        """
        Return a device according to key

        @param key: identifier for a device, either it's id (int) or it's name
        @type key: int or str
        """
        out = None
        try:
            devid = int(key)
        except ValueError:
            for a_dev in self.devices:
                if a_dev.name == key:
                    out = a_dev
        else:
            if len(self.devices) > devid > 0:
                out = self.devices[devid]
        return out


class OpenCL(object):
    """
    Simple class that wraps the structure ocl_tools_extended.h

    This is a static class.
    ocl should be the only instance and shared among all python modules. 
    """
    platforms = []
    nb_devices = 0
    if pyopencl:
        platform = device = pypl = devtype = extensions = pydev = None
        for idx, platform in enumerate(pyopencl.get_platforms()):
            pypl = Platform(platform.name, platform.vendor, platform.version, platform.extensions, idx)
            for idd, device in enumerate(platform.get_devices()):
                ####################################################
                # Nvidia does not report int64 atomics (we are using) ...
                # this is a hack around as any nvidia GPU with double-precision supports int64 atomics
                ####################################################
                extensions = device.extensions
                if (pypl.vendor == "NVIDIA Corporation") and ('cl_khr_fp64' in extensions):
                                extensions += ' cl_khr_int64_base_atomics cl_khr_int64_extended_atomics'
                try:
                    devtype = pyopencl.device_type.to_string(device.type).upper()
                except ValueError:
                    # pocl does not describe itself as a CPU !
                    devtype = "CPU"
                if len(devtype) > 3:
                    devtype = devtype[:3]
                if (pypl.vendor == "NVIDIA Corporation") and (devtype == "GPU") and "compute_capability_major_nv" in dir(device):
                    comput_cap = device.compute_capability_major_nv, device.compute_capability_minor_nv
                    flop_core = NVIDIA_FLOP_PER_CORE.get(comput_cap, min(NVIDIA_FLOP_PER_CORE.values()))
                elif (pypl.vendor == "Advanced Micro Devices, Inc.") and (devtype == "GPU"):
                    flop_core = AMD_FLOP_PER_CORE
                elif devtype == "CPU":
                    flop_core = FLOP_PER_CORE.get(devtype, 1)
                else:
                     flop_core = 1
                workgroup = device.max_work_group_size   
                if (devtype == "CPU") and (pypl.vendor == "Apple"):
                    logger.info("For Apple's OpenCL on CPU: enforce max_work_goup_size=1")
                    workgroup = 1
  
                pydev = Device(device.name, devtype, device.version, device.driver_version, extensions,
                               device.global_mem_size, bool(device.available), device.max_compute_units,
                               device.max_clock_frequency, flop_core, idd, workgroup)
                pypl.add_device(pydev)
                nb_devices += 1
            platforms.append(pypl)
        del platform, device, pypl, devtype, extensions, pydev


    def __repr__(self):
        out = ["OpenCL devices:"]
        for platformid, platform in enumerate(self.platforms):
            out.append("[%s] %s: " % (platformid, platform.name) + ", ".join(["(%s,%s) %s" % (platformid, deviceid, dev.name) for deviceid, dev in enumerate(platform.devices)]))
        return os.linesep.join(out)

    def get_platform(self, key):
        """
        Return a platform according

        @param key: identifier for a platform, either an Id (int) or it's name
        @type key: int or str
        """
        out = None
        try:
            platid = int(key)
        except ValueError:
            for a_plat in self.platforms:
                if a_plat.name == key:
                    out = a_plat
        else:
            if len(self.platforms) > platid > 0:
                out = self.platforms[platid]
        return out

    def select_device(self, dtype="ALL", memory=None, extensions=[], best=True, **kwargs):
        """
        Select a device based on few parameters (at the end, keep the one with most memory)

        @param type: "gpu" or "cpu" or "all" ....
        @param memory: minimum amount of memory (int)
        @param extensions: list of extensions to be present
        @param best: shall we look for the
        """
        if "type" in kwargs:
            dtype = kwargs["type"].upper()
        else:
            dtype = dtype.upper()
        if len(dtype) > 3:
            dtype = dtype[:3]
        best_found = None
        for platformid, platform in enumerate(self.platforms):
            for deviceid, device in enumerate(platform.devices):
                if (dtype in ["ALL", "DEF"]) or (device.type == dtype):
                    if (memory is None) or (memory <= device.memory):
                        found = True
                        for ext in extensions:
                            if ext not in device.extensions:
                                found = False
                        if found:
                            if not best:
                                return platformid, deviceid
                            else:
                                if not best_found:
                                    best_found = platformid, deviceid, device.flops
                                elif best_found[2] < device.flops:
                                    best_found = platformid, deviceid, device.flops
        if best_found:
            return  best_found[0], best_found[1]

    def create_context(self, devicetype="ALL", useFp64=False, platformid=None, deviceid=None):
        """
        Choose a device and initiate a context.

        Devicetypes can be GPU,gpu,CPU,cpu,DEF,ACC,ALL.
        Suggested are GPU,CPU.
        For each setting to work there must be such an OpenCL device and properly installed.
        E.g.: If Nvidia driver is installed, GPU will succeed but CPU will fail. The AMD SDK kit is required for CPU via OpenCL.
        @param devicetype: string in ["cpu","gpu", "all", "acc"]
        @param useFp64: boolean specifying if double precision will be used
        @param platformid: integer
        @param devid: integer
        @return: OpenCL context on the selected device
        """
        if (platformid is not None) and (deviceid is not None):
            platformid = int(platformid)
            deviceid = int(deviceid)
        else:
            if useFp64:
                ids = ocl.select_device(type=devicetype, extensions=["cl_khr_int64_base_atomics"])
            else:
                ids = ocl.select_device(dtype=devicetype)
            if ids:
                platformid = ids[0]
                deviceid = ids[1]
        if (platformid is not None) and  (deviceid is not None):
            ctx = pyopencl.Context(devices=[pyopencl.get_platforms()[platformid].get_devices()[deviceid]])
        else:
            logger.warn("Last chance to get an OpenCL device ... probably not the one requested")
            ctx = pyopencl.create_some_context(interactive=False)
        return ctx


def release_cl_buffers(cl_buffers):
    """
    @param cl_buffer: the buffer you want to release
    @type cl_buffer: dict(str, pyopencl.Buffer)

    This method release the memory of the buffers store in the dict
    """
    for key in cl_buffers:
        if cl_buffers[key] is not None:
            try:
                cl_buffers[key].release()
                cl_buffers[key] = None
            except pyopencl.LogicError:
                logger.error("Error while freeing buffer %s", key)
    return cl_buffers


def allocate_cl_buffers(buffers, device, context):
    """
    @param buffers: the buffers info use to create the pyopencl.Buffer
    @type buffer: list(std, flag, numpy.dtype, int)
    @return: a dict containing the instanciated pyopencl.Buffer
    @rtype: dict(str, pyopencl.Buffer)

    This method instanciate the pyopencl.Buffer from the buffers
    description.
    """
    mem = {}

    # check if enough memory is available on the device
    ualloc = 0
    for _, _, dtype, size in buffers:
        ualloc += numpy.dtype(dtype).itemsize * size
    memory = device.memory
    logger.info("%.3fMB are needed on device which has %.3fMB",
                ualloc / 1.0e6, memory / 1.0e6)
    if ualloc >= memory:
        raise MemoryError("Fatal error in _allocate_buffers. Not enough device memory for buffers (%lu requested, %lu available)" % (ualloc, memory))  # noqa

    # do the allocation
    try:
        for name, flag, dtype, size in buffers:
            mem[name] = \
                pyopencl.Buffer(context, flag,
                                numpy.dtype(dtype).itemsize * size)
    except pyopencl.MemoryError as error:
        release_cl_buffers(mem)
        raise MemoryError(error)

    return mem

if pyopencl:
    ocl = OpenCL()
    if ocl.nb_devices == 0:
        ocl = None
else:
    ocl = None




"""
Module Sift for calculating SIFT keypoint using PyOpenCL
"""
version = "0.3.0"
import os
sift_home = os.path.dirname(os.path.abspath(__file__))
import sys, logging
logging.basicConfig()
from .plan import SiftPlan
from .match import MatchPlan
from .alignment import LinearAlign




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#

"""
Test suite for transformation kernel
"""

from __future__ import division

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-05-28"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""

import time, os, logging
import numpy
import pyopencl, pyopencl.array
import scipy, scipy.misc, scipy.ndimage, pylab
import sys
import unittest
from utilstest import UtilsTest, getLogger, ctx
from test_image_functions import * #for Python implementation of tested functions
from test_image_setup import *
import sift_pyocl as sift
from sift_pyocl.utils import calc_size
logger = getLogger(__file__)
if logger.getEffectiveLevel() <= logging.INFO:
    PROFILE = True
    queue = pyopencl.CommandQueue(ctx, properties=pyopencl.command_queue_properties.PROFILING_ENABLE)
    import pylab
else:
    PROFILE = False
    queue = pyopencl.CommandQueue(ctx)

SHOW_FIGURES = False
IMAGE_RESHAPE = True
USE_LENA = False

print "working on %s" % ctx.devices[0].name





class test_transform(unittest.TestCase):
    def setUp(self):
        
        kernel_path = os.path.join(os.path.dirname(os.path.abspath(sift.__file__)), "transform.cl")
        kernel_src = open(kernel_path).read()
        self.program = pyopencl.Program(ctx, kernel_src).build() #.build('-D WORKGROUP_SIZE=%s' % wg_size)
        self.wg = (1, 128)



    def tearDown(self):
        self.program = None
        

    
    
    def image_reshape(self,img,output_height,output_width,image_height, image_width):
        '''
        Reshape the image to get a bigger image with the input image in the center
        
        '''
        image3 = numpy.zeros((output_height,output_width),dtype=numpy.float32)
        d1 = (output_width - image_width)/2
        d0 = (output_height - image_height)/2
        image3[d0:-d0,d1:-d1] = numpy.copy(img)
        image = image3
        image_height, image_width = output_height, output_width
        return image, image_height, image_width
    
    
    
    
    
    
    
    def matching_correction(self,image,image2):
        '''
        Computes keypoints for two images and try to align image2 on image1
        '''
        #computing keypoints matching
        s = sift.SiftPlan(template=image,devicetype="gpu")
        kp1 = s.keypoints(image)
        kp2 = s.keypoints(image2) #image2 and image must have the same size
        m = sift.MatchPlan(devicetype="GPU")
        matching = m.match(kp2,kp1)
        N = matching.shape[0]
        #solving normals equations for least square fit
        X = numpy.zeros((2*N,6))
        X[::2,2:] = 1,0,0,0
        X[::2,0] = matching.x[:,0]
        X[::2,1] = matching.y[:,0]
        X[1::2,0:3] = 0,0,0
        X[1::2,3] = matching.x[:,0]
        X[1::2,4] = matching.y[:,0]
        X[1::2,5] = 1
        y = numpy.zeros((2*N,1))
        y[::2,0] = matching.x[:,1]
        y[1::2,0] = matching.y[:,1]
        #A = numpy.dot(X.transpose(),X)
        #sol = numpy.dot(numpy.linalg.inv(A),numpy.dot(X.transpose(),y))
        sol = numpy.dot(numpy.linalg.pinv(X),y)
        MSE = numpy.linalg.norm(y - numpy.dot(X,sol))**2/N #value of the sum of residuals at "sol"
        return sol, MSE






    def test_transform(self):
        '''
        tests transform kernel
        '''    



        if (USE_LENA):
            #original image
            image = scipy.misc.lena().astype(numpy.float32)
            image = numpy.ascontiguousarray(image[0:512,0:512])
            image_height, image_width = image.shape
            #transformation
            angle = 1.9 #numpy.pi/5.0
    #        matrix = numpy.array([[numpy.cos(angle),-numpy.sin(angle)],[numpy.sin(angle),numpy.cos(angle)]],dtype=numpy.float32)
    #        offset_value = numpy.array([1000.0, 100.0],dtype=numpy.float32)
    #        matrix = numpy.array([[0.9,0.2],[-0.4,0.9]],dtype=numpy.float32)
    #        offset_value = numpy.array([-20.0,256.0],dtype=numpy.float32)
            matrix = numpy.array([[1.0,-0.75],[0.7,0.5]],dtype=numpy.float32)
            
            offset_value = numpy.array([250.0, -150.0],dtype=numpy.float32)
           
            image2 = scipy.ndimage.interpolation.affine_transform(image,matrix,offset=offset_value,order=1, mode="constant")
        
        else: #use images of a stack
            image = scipy.misc.imread("/home/paleo/Titanium/test/frame0.png")
            image2 = scipy.misc.imread("/home/paleo/Titanium/test/frame1.png")
            offset_value = numpy.array([0.0, 0.0],dtype=numpy.float32)
            image_height, image_width = image.shape
            image2_height, image2_width = image2.shape
            
        fill_value = numpy.float32(0.0)
        mode = numpy.int32(1)   
            
        if IMAGE_RESHAPE: #turns out that image should always be reshaped
            output_height, output_width = int(3000), int(3000)
            image, image_height, image_width = self.image_reshape(image,output_height,output_width,image_height,image_width)
            image2, image2_height, image2_width = self.image_reshape(image2,output_height,output_width,image2_height,image2_width) 
            
            
       
        else: output_height, output_width = int(image_height*numpy.sqrt(2)),int(image_width*numpy.sqrt(2))
        print "Image : (%s, %s) -- Output: (%s, %s)" %(image_height, image_width , output_height, output_width)
        
        
        
            
        
        
        
        
        
        
        
        #perform correction by least square
        sol, MSE = self.matching_correction(image,image2)
        print sol
        
        
        correction_matrix = numpy.zeros((2,2),dtype=numpy.float32)
        correction_matrix[0] = sol[0:2,0]
        correction_matrix[1] = sol[3:5,0]
        matrix_for_gpu = correction_matrix.reshape(4,1) #for float4 struct
        offset_value[0] = sol[2,0]
        offset_value[1] = sol[5,0]
        
        wg = 8,8
        shape = calc_size((output_width,output_height), wg)
        gpu_image = pyopencl.array.to_device(queue, image2)
        gpu_output = pyopencl.array.empty(queue, (output_height, output_width), dtype=numpy.float32, order="C")
        gpu_matrix = pyopencl.array.to_device(queue,matrix_for_gpu)
        gpu_offset = pyopencl.array.to_device(queue,offset_value)
        image_height, image_width = numpy.int32((image_height, image_width))
        output_height, output_width = numpy.int32((output_height, output_width))
        
        t0 = time.time()
        k1 = self.program.transform(queue, shape, wg,
                gpu_image.data, gpu_output.data, gpu_matrix.data, gpu_offset.data, 
                image_width, image_height, output_width, output_height, fill_value, mode)
        res = gpu_output.get()
        t1 = time.time()
#        print res[0,0]
        
        ref = scipy.ndimage.interpolation.affine_transform(image2,correction_matrix,
            offset=offset_value, output_shape=(output_height,output_width),order=1, mode="constant", cval=fill_value)
        t2 = time.time()
        
        delta = abs(res-image)
        delta_arg = delta.argmax()
        delta_max = delta.max()
#        delta_mse_res = ((res-image)**2).sum()/image.size
#        delta_mse_ref = ((ref-image)**2).sum()/image.size
        at_0, at_1 = delta_arg/output_width, delta_arg%output_width
        print("Max error: %f at (%d, %d)" %(delta_max, at_0, at_1))
#        print("Mean Squared Error Res/Original : %f" %(delta_mse_res))
#        print("Mean Squared Error Ref/Original: %f" %(delta_mse_ref))
        print("minimal MSE according to least squares : %f" %MSE)
#        print res[at_0,at_1]
#        print ref[at_0,at_1]
        
        SHOW_FIGURES = True
        if SHOW_FIGURES:
            fig = pylab.figure()
            sp1 = fig.add_subplot(221,title="Input image")
            sp1.imshow(image, interpolation="nearest")
            sp2 = fig.add_subplot(222,title="Image after deformation")
            sp2.imshow(image2, interpolation="nearest")
            sp2 = fig.add_subplot(223,title="Corrected image (OpenCL)")
            sp2.imshow(res, interpolation="nearest")
            sp2 = fig.add_subplot(224,title="Corrected image (Scipy)")
            sp2.imshow(ref, interpolation="nearest")
#            sp2.imshow(ref, interpolation="nearest")
#            sp3 = fig.add_subplot(223,title="delta (max = %f)" %delta_max)
#            sh3 = sp3.imshow(delta[:,:], interpolation="nearest")
#            cbar = fig.colorbar(sh3)
            fig.show()
            raw_input("enter")


        if PROFILE:
            logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
            logger.info("Transformation took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))
            










            
            
            

def test_suite_transform():
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_transform("test_transform"))
    return testSuite

if __name__ == '__main__':
    mysuite = test_suite_transform()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        sys.exit(1)




#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#

"""
Test suite for all preprocessing kernels.
"""

from __future__ import division

__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2013-05-28"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""

import time, os, logging
import numpy
import pyopencl, pyopencl.array
import scipy, scipy.misc
import sys
import unittest
from utilstest import UtilsTest, getLogger, ctx
import sift_pyocl as sift
from sift_pyocl.utils import calc_size
import math

logger = getLogger(__file__)

if logger.getEffectiveLevel() <= logging.INFO:
    PROFILE = True
    queue = pyopencl.CommandQueue(ctx, properties=pyopencl.command_queue_properties.PROFILING_ENABLE)
    import pylab
else:
    PROFILE = False
    queue = pyopencl.CommandQueue(ctx)

def normalize(img, max_out=255):
    """
    Numpy implementation of the normalization
    """
    fimg = img.astype("float32")
    img_max = fimg.max()
    img_min = fimg.min()
    return max_out * (fimg - img_min) / (img_max - img_min)

def shrink(img, xs, ys):
    return img[0::ys, 0::xs]
def shrink_cython(img, xs, ys):
    try:
        import feature
    except:
        return img[0::ys, 0::xs]
    else:
        return feature.shrink(img, xs)


def binning(input_img, binsize):
    """
    @param input_img: input ndarray
    @param binsize: int or 2-tuple representing the size of the binning
    @return: binned input ndarray

    TODO: Not used here
    """
    inputSize = input_img.shape
    outputSize = []
    assert(len(inputSize) == 2)
    if isinstance(binsize, int):
        binsize = (binsize, binsize)
    outputSize = [int(math.ceil(float(i) / j)) for i, j in zip(inputSize, binsize)]
    bigSize = [i * j  for i, j in zip(outputSize, binsize)]
    delta = [j - i  for i, j in zip(inputSize, bigSize)]
    big_array = numpy.empty(bigSize, input_img.dtype)
    big_array[:inputSize[0],:inputSize[1]]=input_img
    #corner
    big_array[inputSize[0]:, inputSize[1]:] = input_img[-1:-delta[0] - 1:-1, -1:-delta[1] - 1:-1]
    # 2 sides
    big_array[inputSize[0]:, :inputSize[1]] = input_img[-1:-delta[0] - 1:-1, :]
    big_array[:inputSize[0], inputSize[1]:] = input_img[:, -1:-delta[1] - 1:-1]

    if numpy.array(binsize).prod() < 50:
        out = numpy.zeros(tuple(outputSize))
#        print input_img.shape, out.shape, big_array.shape, binsize
        for i in xrange(binsize[0]):
            for j in xrange(binsize[1]):
                out += big_array[i::binsize[0], j::binsize[1]]
    else:
        temp = big_array.copy()
        temp.shape = (outputSize[0], binsize[0], outputSize[1], binsize[1])
        out = temp.sum(axis=3).sum(axis=1)
    return out

class test_preproc(unittest.TestCase):
    def setUp(self):
        self.input = numpy.ascontiguousarray(scipy.misc.lena()[:510, :511])
        self.gpudata = pyopencl.array.empty(queue, self.input.shape, dtype=numpy.float32, order="C")
        kernel_path = os.path.join(os.path.dirname(os.path.abspath(sift.__file__)), "preprocess.cl")
        reduct_path = os.path.join(os.path.dirname(os.path.abspath(sift.__file__)), "reductions.cl")
        kernel_src = open(kernel_path).read()
        reduct_src = open(reduct_path).read()
        self.program = pyopencl.Program(ctx, kernel_src).build()
        self.reduction = pyopencl.Program(ctx, reduct_src).build()
        self.IMAGE_W = numpy.int32(self.input.shape[-1])
        self.IMAGE_H = numpy.int32(self.input.shape[0])
        self.wg = (32, 16)#(256, 2) #(32, 16) # (2, 256)
        self.shape = calc_size((self.IMAGE_W, self.IMAGE_H), self.wg)
#        print self.shape
        self.binning = (4, 2) # Nota if wg < ouptup size weired results are expected !
#        self.binning = (2, 2)
        self.red_size = 128 #reduction size
        self.twofivefive = pyopencl.array.to_device(queue, numpy.array([255], numpy.float32))
        self.buffers_max_min = pyopencl.array.empty(queue, (self.red_size, 2), dtype=numpy.float32)  # temporary buffer for max/min reduction
        self.buffers_min = pyopencl.array.empty(queue, (1), dtype=numpy.float32)
        self.buffers_max = pyopencl.array.empty(queue, (1), dtype=numpy.float32)

    def tearDown(self):
        self.input = None
        self.program = None
        self.twofivefive = None
        self.buffers_max_min = None
        self.buffers_max = None
        self.buffers_min = None

    def test_uint8(self):
        """
        tests the uint8 kernel
        """
        lint = self.input.astype(numpy.uint8)
        t0 = time.time()
        au8 = pyopencl.array.to_device(queue, lint)
        k1 = self.program.u8_to_float(queue, self.shape, self.wg, au8.data, self.gpudata.data, self.IMAGE_W, self.IMAGE_H)
#        print abs(au8.get() - self.gpudata.get()).max()
        k2 = self.reduction.max_min_global_stage1(queue, (self.red_size * self.red_size,), (self.red_size,),
                                                               self.gpudata.data,
                                                               self.buffers_max_min.data,
                                                               (self.IMAGE_W * self.IMAGE_H))
        k3 = self.reduction.max_min_global_stage2(queue, (self.red_size,), (self.red_size,),
                                                               self.buffers_max_min.data,
                                                               self.buffers_max.data,
                                                               self.buffers_min.data)
#        print self.buffers_max.get(), self.buffers_min.get(), self.input.min(), self.input.max()
        k4 = self.program.normalizes(queue, self.shape, self.wg,
                                     self.gpudata.data,
                                     self.buffers_min.data,
                                     self.buffers_max.data,
                                     self.twofivefive.data,
                                     self.IMAGE_W, self.IMAGE_H)
        k4.wait()
        res = self.gpudata.get()
        t1 = time.time()
        ref = normalize(lint)
        t2 = time.time()
        delta = abs(ref - res).max()
        if PROFILE:
            logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
            logger.info("Conversion uint8->float took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))
            logger.info("Reduction stage1 took        %.3fms" % (1e-6 * (k2.profile.end - k2.profile.start)))
            logger.info("Reduction stage2 took        %.3fms" % (1e-6 * (k3.profile.end - k3.profile.start)))
            logger.info("Normalization                %.3fms" % (1e-6 * (k4.profile.end - k4.profile.start)))
            logger.info("--------------------------------------")

        self.assert_(delta < 1e-4, "delta=%s" % delta)

    def test_uint16(self):
        """
        tests the uint16 kernel
        """
        lint = self.input.astype(numpy.uint16)
        t0 = time.time()
        au8 = pyopencl.array.to_device(queue, lint)
        k1 = self.program.u16_to_float(queue, self.shape, self.wg, au8.data, self.gpudata.data, self.IMAGE_W, self.IMAGE_H)
        k2 = self.reduction.max_min_global_stage1(queue, (self.red_size * self.red_size,), (self.red_size,),
                                                               self.gpudata.data,
                                                               self.buffers_max_min.data,
                                                               (self.IMAGE_W * self.IMAGE_H))
        k3 = self.reduction.max_min_global_stage2(queue, (self.red_size,), (self.red_size,),
                                                               self.buffers_max_min.data,
                                                               self.buffers_max.data,
                                                               self.buffers_min.data)
        k4 = self.program.normalizes(queue, self.shape, self.wg,
                                     self.gpudata.data,
                                     self.buffers_min.data,
                                     self.buffers_max.data,
                                     self.twofivefive.data,
                                     self.IMAGE_W, self.IMAGE_H)
        k4.wait()
        res = self.gpudata.get()
        t1 = time.time()
        ref = normalize(lint)
        t2 = time.time()
        delta = abs(ref - res).max()
        if PROFILE:
            logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
            logger.info("Conversion uint16->float took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))
            logger.info("Reduction stage1 took         %.3fms" % (1e-6 * (k2.profile.end - k2.profile.start)))
            logger.info("Reduction stage2 took         %.3fms" % (1e-6 * (k3.profile.end - k3.profile.start)))
            logger.info("Normalization                 %.3fms" % (1e-6 * (k4.profile.end - k4.profile.start)))
            logger.info("--------------------------------------")
        self.assert_(delta < 1e-4, "delta=%s" % delta)

    def test_int32(self):
        """
        tests the int32 kernel
        """
        lint = self.input.astype(numpy.int32)
        t0 = time.time()
        au8 = pyopencl.array.to_device(queue, lint)
        k1 = self.program.s32_to_float(queue, self.shape, self.wg, au8.data, self.gpudata.data, self.IMAGE_W, self.IMAGE_H)
        k2 = self.reduction.max_min_global_stage1(queue, (self.red_size * self.red_size,), (self.red_size,),
                                                               self.gpudata.data,
                                                               self.buffers_max_min.data,
                                                               (self.IMAGE_W * self.IMAGE_H))
        k3 = self.reduction.max_min_global_stage2(queue, (self.red_size,), (self.red_size,),
                                                               self.buffers_max_min.data,
                                                               self.buffers_max.data,
                                                               self.buffers_min.data)
        k4 = self.program.normalizes(queue, self.shape, self.wg,
                                     self.gpudata.data,
                                     self.buffers_min.data,
                                     self.buffers_max.data,
                                     self.twofivefive.data,
                                     self.IMAGE_W, self.IMAGE_H)
        k4.wait()
        res = self.gpudata.get()
        t1 = time.time()
        ref = normalize(lint)
        t2 = time.time()
        delta = abs(ref - res).max()
        if PROFILE:
            logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
            logger.info("Conversion int32->float took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))
            logger.info("Reduction stage1 took        %.3fms" % (1e-6 * (k2.profile.end - k2.profile.start)))
            logger.info("Reduction stage2 took        %.3fms" % (1e-6 * (k3.profile.end - k3.profile.start)))
            logger.info("Normalization                %.3fms" % (1e-6 * (k4.profile.end - k4.profile.start)))
            logger.info("--------------------------------------")
        self.assert_(delta < 1e-4, "delta=%s" % delta)

    def test_int64(self):
        """
        tests the int64 kernel
        """
        lint = self.input.astype(numpy.int64)
        t0 = time.time()
        au8 = pyopencl.array.to_device(queue, lint)
        k1 = self.program.s64_to_float(queue, self.shape, self.wg, au8.data, self.gpudata.data, self.IMAGE_W, self.IMAGE_H)
        k2 = self.reduction.max_min_global_stage1(queue, (self.red_size * self.red_size,), (self.red_size,),
                                                               self.gpudata.data,
                                                               self.buffers_max_min.data,
                                                               (self.IMAGE_W * self.IMAGE_H))
        k3 = self.reduction.max_min_global_stage2(queue, (self.red_size,), (self.red_size,),
                                                               self.buffers_max_min.data,
                                                               self.buffers_max.data,
                                                               self.buffers_min.data)
        k4 = self.program.normalizes(queue, self.shape, self.wg,
                                     self.gpudata.data,
                                     self.buffers_min.data,
                                     self.buffers_max.data,
                                     self.twofivefive.data,
                                     self.IMAGE_W, self.IMAGE_H)
        res = self.gpudata.get()
        t1 = time.time()
        ref = normalize(lint)
        t2 = time.time()
        delta = abs(ref - res).max()
        if PROFILE:
            logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
            logger.info("Conversion int64->float took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))
            logger.info("Reduction stage1 took        %.3fms" % (1e-6 * (k2.profile.end - k2.profile.start)))
            logger.info("Reduction stage2 took        %.3fms" % (1e-6 * (k3.profile.end - k3.profile.start)))
            logger.info("Normalization                %.3fms" % (1e-6 * (k4.profile.end - k4.profile.start)))
            logger.info("--------------------------------------")
        self.assert_(delta < 1e-4, "delta=%s" % delta)

    def test_rgb(self):
        """
        tests the int64 kernel
        """
        lint = numpy.empty((self.input.shape[0], self.input.shape[1], 3), dtype=numpy.uint8)
        lint[:, :, 0] = self.input.astype(numpy.uint8)
        lint[:, :, 1] = self.input.astype(numpy.uint8)
        lint[:, :, 2] = self.input.astype(numpy.uint8)
        t0 = time.time()
        au8 = pyopencl.array.to_device(queue, lint)
        k1 = self.program.rgb_to_float(queue, self.shape, self.wg, au8.data, self.gpudata.data, self.IMAGE_W, self.IMAGE_H)
        k2 = self.reduction.max_min_global_stage1(queue, (self.red_size * self.red_size,), (self.red_size,),
                                                               self.gpudata.data,
                                                               self.buffers_max_min.data,
                                                               (self.IMAGE_W * self.IMAGE_H))
        k3 = self.reduction.max_min_global_stage2(queue, (self.red_size,), (self.red_size,),
                                                               self.buffers_max_min.data,
                                                               self.buffers_max.data,
                                                               self.buffers_min.data)
        k4 = self.program.normalizes(queue, self.shape, self.wg,
                                     self.gpudata.data,
                                     self.buffers_min.data,
                                     self.buffers_max.data,
                                     self.twofivefive.data,
                                     self.IMAGE_W, self.IMAGE_H)
        res = self.gpudata.get()
        t1 = time.time()
        ref = normalize(lint.max(axis= -1))
        t2 = time.time()
        delta = abs(ref - res).max()
        if PROFILE:
            logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
            logger.info("Conversion  RGB ->float took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))
            logger.info("Reduction stage1 took        %.3fms" % (1e-6 * (k2.profile.end - k2.profile.start)))
            logger.info("Reduction stage2 took        %.3fms" % (1e-6 * (k3.profile.end - k3.profile.start)))
            logger.info("Normalization                %.3fms" % (1e-6 * (k4.profile.end - k4.profile.start)))
            logger.info("--------------------------------------")
        self.assert_(delta < 1e-4, "delta=%s" % delta)


    def test_shrink(self):
        """
        Test shrinking kernel
        """
        lint = self.input.astype(numpy.float32)
        out_shape = tuple(int(math.ceil(float(i) / j)) for i, j in zip((self.IMAGE_H, self.IMAGE_W), self.binning))
        t0 = time.time()
        inp_gpu = pyopencl.array.to_device(queue, lint)
        out_gpu = pyopencl.array.empty(queue, out_shape, dtype=numpy.float32, order="C")
        k1 = self.program.shrink(queue, calc_size((out_shape[1], out_shape[0]), self.wg), self.wg,
                                 inp_gpu.data, out_gpu.data,
                                 numpy.int32(self.binning[1]), numpy.int32(self.binning[0]),
                                 self.IMAGE_W, self.IMAGE_H,
                                 numpy.int32(out_shape[1]), numpy.int32(out_shape[0]))
        res = out_gpu.get()
        t1 = time.time()
        ref = shrink(lint, xs=self.binning[1], ys=self.binning[0])
        t2 = time.time()
#        print ref.shape, res.shape
        delta = abs(ref - res).max()
        if PROFILE:
            logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
            logger.info("Shrinking  took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))
            fig = pylab.figure()
            fig.suptitle('Shrinking by %s,%s' % self.binning)
            sp1 = fig.add_subplot(221)
            sp1.imshow(lint, interpolation="nearest")
            sp1.set_title("Input")
            sp2 = fig.add_subplot(222)
            sp2.imshow(ref, interpolation="nearest")
            sp2.set_title("Reference")
            sp3 = fig.add_subplot(223)
            sp3.imshow(ref - res, interpolation="nearest")
            sp3.set_title("Delta= %s" % delta)
            sp4 = fig.add_subplot(224)
            sp4.imshow(res, interpolation="nearest")
            sp4.set_title("GPU")
            fig.show()
            raw_input("enter")
        self.assert_(delta < 1e-6, "delta=%s" % delta)


    def test_bin(self):
        """
        Test binning kernel
        """
        lint = numpy.ascontiguousarray(self.input, numpy.float32)

        out_shape = tuple(int(math.ceil((float(i) / j))) for i, j in zip(self.input.shape, self.binning))
        t0 = time.time()
        inp_gpu = pyopencl.array.to_device(queue, lint)
        out_gpu = pyopencl.array.empty(queue, out_shape, dtype=numpy.float32, order="C")
        k1 = self.program.bin(queue, calc_size((out_shape[1], out_shape[0]), self.wg), self.wg, inp_gpu.data, out_gpu.data,
                                 numpy.int32(self.binning[1]), numpy.int32(self.binning[0]),
                                 numpy.int32(lint.shape[1]), numpy.int32(lint.shape[0]),
                                 numpy.int32(out_shape[1]), numpy.int32(out_shape[0]))
        res = out_gpu.get()
        t1 = time.time()
        ref = binning(lint, self.binning) / self.binning[0] / self.binning[1]
        t2 = time.time()
#        print ref.shape, res.shape
        delta = abs(ref - res).max()
        if PROFILE:
            logger.info("Global execution time: CPU %.3fms, GPU: %.3fms." % (1000.0 * (t2 - t1), 1000.0 * (t1 - t0)))
            logger.info("Binning took %.3fms" % (1e-6 * (k1.profile.end - k1.profile.start)))
            fig = pylab.figure()
            fig.suptitle('Binning by %s,%s' % self.binning)
            sp1 = fig.add_subplot(221)
            sp1.imshow(lint, interpolation="nearest")
            sp1.set_title("Input")
            sp2 = fig.add_subplot(222)
            sp2.imshow(ref, interpolation="nearest")
            sp2.set_title("Reference")
            sp3 = fig.add_subplot(223)
            sp3.imshow(ref - res, interpolation="nearest")
            sp3.set_title("Delta= %s" % delta)
            sp4 = fig.add_subplot(224)
            sp4.imshow(res, interpolation="nearest")
            sp4.set_title("GPU")
            fig.show()
            raw_input("enter")
        self.assert_(delta < 1e-6, "delta=%s" % delta)



def test_suite_preproc():
    testSuite = unittest.TestSuite()
    testSuite.addTest(test_preproc("test_uint8"))
    testSuite.addTest(test_preproc("test_uint16"))
    testSuite.addTest(test_preproc("test_int32"))
    testSuite.addTest(test_preproc("test_int64"))
    testSuite.addTest(test_preproc("test_rgb"))
    testSuite.addTest(test_preproc("test_shrink"))
    testSuite.addTest(test_preproc("test_bin"))
    return testSuite

if __name__ == '__main__':
    mysuite = test_suite_preproc()
    runner = unittest.TextTestRunner()
    if not runner.run(mysuite).wasSuccessful():
        sys.exit(1)





#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
#    Project: Sift implementation in Python + OpenCL
#             https://github.com/kif/sift_pyocl
#

"""
Installer script for SIFT algorithm in PyOpenCL 
"""

from __future__ import division, with_statement, print_function


__authors__ = ["Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "2014-10-28"
__status__ = "stable"
__license__ = """
Permission is hereby granted, free of charge, to any person
obtaining a copy of ethis software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

"""



import os, sys, glob, shutil, ConfigParser, platform
from distutils.core import setup, Extension, Command
from numpy.distutils.misc_util import get_numpy_include_dirs
from distutils.command.install_data import install_data
from distutils.sysconfig import get_python_lib

# ###############################################################################
# Check for Cython
# ###############################################################################
try:
    from Cython.Distutils import build_ext
    CYTHON = True
except ImportError:
    CYTHON = False
if CYTHON:
    try:
        import Cython.Compiler.Version
    except ImportError:
        CYTHON = False
    else:
        if Cython.Compiler.Version.version < "0.17":
            CYTHON = False

if CYTHON:
    cython_c_ext = ".pyx"
else:
    cython_c_ext = ".c"
    from distutils.command.build_ext import build_ext
cmdclass = {}


def rewriteManifest(with_testimages=False):
    """
    Rewrite the "Manifest" file ... if needed

    @param with_testimages: include
    """
    base = os.path.dirname(os.path.abspath(__file__))
    manifest_file = os.path.join(base, "MANIFEST.in")
    if not os.path.isfile(manifest_file):
        print("MANIFEST file is missing !!!")
        return
    manifest = [i.strip() for i in open(manifest_file)]
    changed = False

    if with_testimages:
        testimages = ["test/testimages/" + i for i in os.listdir(os.path.join(base, "test", "testimages"))]
        for image in testimages:
            if image not in manifest:
                manifest.append("include " + image)
                changed = True
    else:
        for line in manifest[:]:
            if line.startswith("include test/testimages"):
                changed = True
                manifest.remove(line)
    if changed:
        with open(manifest_file, "w") as f:
            f.write(os.linesep.join(manifest))
        # remove MANIFEST: will be re generated !
        os.unlink(manifest_file[:-3])

if ("sdist" in sys.argv):
    if ("--with-testimages" in sys.argv):
        sys.argv.remove("--with-testimages")
        rewriteManifest(with_testimages=True)
    else:
        rewriteManifest(with_testimages=False)


installDir = "sift" #relative to site-packages ...
data_files = [(installDir, glob.glob("openCL/*.cl"))]


class smart_install_data(install_data):
    def run(self):
        install_cmd = self.get_finalized_command('install')
        self.install_dir = getattr(install_cmd, 'install_lib')
        print("DATA to be installed in %s" % self.install_dir)
        return install_data.run(self)
cmdclass['install_data'] = smart_install_data


if sys.platform == "win32":
    # This is for mingw32/gomp?
#    data_files[0][1].append(os.path.join("dll", "pthreadGC2.dll"))
    root = os.path.dirname(os.path.abspath(__file__))
    tocopy_files = []
    script_files = []
    for i in os.listdir(os.path.join(root, "scripts")):
        if os.path.isfile(os.path.join(root, "scripts", i)):
            if i.endswith(".py"):
                script_files.append(os.path.join("scripts", i))
            else:
                tocopy_files.append(os.path.join("scripts", i))
    for i in tocopy_files:
        filein = os.path.join(root, i)
        if (filein + ".py") not in script_files:
            shutil.copyfile(filein, filein + ".py")
            script_files.append(filein + ".py")

else:
    script_files = glob.glob("scripts/*")

version = [eval(l.split("=")[1]) for l in open(os.path.join(os.path.dirname(
    os.path.abspath(__file__)), "sift-src", "__init__.py"))
    if l.strip().startswith("version")][0]


# We subclass the build_ext class in order to handle compiler flags
# for openmp and opencl etc in a cross platform way
translator = {
        # Compiler
            # name, compileflag, linkflag
        'msvc' : {
            'openmp' : ('/openmp', ' '),
            'debug'  : ('/Zi', ' '),
            'OpenCL' : 'OpenCL',
            },
        'mingw32':{
            'openmp' : ('-fopenmp', '-fopenmp'),
            'debug'  : ('-g', '-g'),
            'stdc++' : 'stdc++',
            'OpenCL' : 'OpenCL'
            },
        'default':{
            'openmp' : ('-fopenmp', '-fopenmp'),
            'debug'  : ('-g', '-g'),
            'stdc++' : 'stdc++',
            'OpenCL' : 'OpenCL'
            }
        }


class build_ext_sift(build_ext):
    def build_extensions(self):
        if self.compiler.compiler_type in translator:
            trans = translator[self.compiler.compiler_type]
        else:
            trans = translator['default']

        for e in self.extensions:
            e.extra_compile_args = [ trans[a][0] if a in trans else a
                                    for a in e.extra_compile_args]
            e.extra_link_args = [ trans[a][1] if a in trans else a
                                 for a in e.extra_link_args]
            e.libraries = filter(None, [ trans[a] if a in trans else None
                                        for a in e.libraries])

            # If you are confused look here:
            # print e, e.libraries
            # print e.extra_compile_args
            # print e.extra_link_args
        build_ext.build_extensions(self)
cmdclass['build_ext'] = build_ext_sift

class PyTest(Command):
    user_options = []
    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        import sys, subprocess
        os.chdir("test")
        errno = subprocess.call([sys.executable, 'test_all.py'])
        if errno != 0:
            print("Tests did not pass !!!")
            #raise SystemExit(errno)
        else:
            print("All Tests passed")
        os.chdir("..")
cmdclass['test'] = PyTest

#######################
# build_doc commandes #
#######################

try:
    import sphinx
    import sphinx.util.console
    sphinx.util.console.color_terminal = lambda: False
    from sphinx.setup_command import BuildDoc
except ImportError:
    sphinx = None

if sphinx:
    class build_doc(BuildDoc):

        def run(self):

            # make sure the python path is pointing to the newly built
            # code so that the documentation is built on this and not a
            # previously installed version

            build = self.get_finalized_command('build')
            sys.path.insert(0, os.path.abspath(build.build_lib))

            # Build the Users Guide in HTML and TeX format
            for builder in ('html', 'latex'):
                self.builder = builder
                self.builder_target_dir = os.path.join(self.build_dir, builder)
                self.mkpath(self.builder_target_dir)
                builder_index = 'index_{0}.txt'.format(builder)
                BuildDoc.run(self)
            sys.path.pop(0)
    cmdclass['build_doc'] = build_doc

classifiers = """\
Development Status :: 5 - Production/Stable
Intended Audience :: Developers
Programming Language :: Python
Topic :: Software Development :: Libraries :: Python Modules
Operating System :: Microsoft :: Windows
Operating System :: Unix
Operating System :: MacOS :: MacOS X
Operating System :: POSIX

"""


setup(name='sift_pyocl',
      version=version,
      author="Pierre Paleo, Jérôme Kieffer",
      author_email="jerome.kieffer@esrf.fr",
      description='Python/OpenCL implementation of Sift algorithm image alignment',
      url="https://github.com/kif/sift_pyocl",
      download_url="https://github.com/kif/sift_pyocl/archive/master.zip",
      scripts=script_files,
#      ext_package="sift",
#      ext_modules=[Extension(**dico) for dico in ext_modules],
      packages=["sift_pyocl"],
      package_dir={"sift_pyocl": "sift-src" },
      test_suite="test",
      cmdclass=cmdclass,
      data_files=data_files,
      classifiers=filter(None, classifiers.split("\n")),
      license="MIT"
      )



#print(data_files)
try:
    import pyopencl
except ImportError:
    print("""sift can use pyopencl to run on parallel accelerators like GPU; this is not optional !!!.
This python module can be found on:
http://pypi.python.org/pypi/pyopencl
""")



